!code to take in radius,latitude, and longitude values
!from command line, and return absolute velocity values
!based on an A3d format perturbation model
!Assumes mantle model: radius values less than 3000 km are
!interpreted as depths.
!------------------------------------------------------
program A3d_absolute
!------------------------------------------------------

use module_A3d
use def_gparam
use f2kcli !command line interface
implicit none

character(len=20) :: arg
real(DP) :: rad,lat,lon,theta,phi,vs,vp,rho,Qs,xi,fi,eta
integer :: ifanis

if(command_argument_count()/=3.and.command_argument_count()/=4) &
     stop 'usage: A3d_absolute rad lat lon ifanis'


ifanis=0
call get_command_argument(1,value=arg)
read(arg,*) rad
call get_command_argument(2,value=arg)
read(arg,*) lat
call get_command_argument(3,value=arg)
read(arg,*) lon
if(command_argument_count()==4) then
   call get_command_argument(4,value=arg)
   read(arg,*) ifanis
endif

!convert to radians
theta=deg2rad*(90.0-lat)
phi=deg2rad*lon

!convert depth to radius if necessary
if (rad<3000.d0) then
   print*,'Interpreting radius value ',rad,' as depth (km).'
   rad=6371.0-rad
else if (rad>6371.0) then
   print*,'Radius ',rad,' is outside the Earth. Make sure units are km.'
   stop
endif

!read model
call init_A3d(0)

!get model values
if(ifanis==0) then
   call A3d_full(rad,theta,phi,ifanis,vs,vp,rho,Qs)
else
   call A3d_full(rad,theta,phi,ifanis,vs,vp,rho,Qs,xi,fi,eta)
endif

!output values to stdout
if(ifanis==0) then
   print*,vs,vp,rho,Qs
else
   print*,vs,vp,rho,xi,fi,eta,Qs
endif

end program A3d_absolute


