!code to take in radius,latitude, and longitude values
!from command line, and return relative velocity values
!based on an A3d format perturbation model
!Assumes mantle model: radius values less than 3000 km are
!interpreted as depths.
!------------------------------------------------------
program A3d_relative
!------------------------------------------------------

use module_A3d
use def_gparam
use f2kcli !command line interface
implicit none

character(len=20) :: arg
character :: par
real(DP) :: rad,lat,lon,theta,phi
real(DP), dimension(:), allocatable :: dv
integer :: npar,i

if(command_argument_count()/=3) &
     stop 'usage: A3d_relative rad lat lon'


call get_command_argument(1,value=arg)
read(arg,*) rad
call get_command_argument(2,value=arg)
read(arg,*) lat
call get_command_argument(3,value=arg)
read(arg,*) lon

!convert to radians
theta=deg2rad*(90.0-lat)
phi=deg2rad*lon

!convert depth to radius as necessary
if (rad<3000.d0) then
   print*,'Interpreting radius value ',rad,' as depth (km).'
   rad=6371.0-rad
else if (rad>6371.0) then
   print*,'Radius ',rad,' is outside the Earth. Make sure units are km.'
   stop
endif

!read model
call init_A3d(0)

!determine par number and get model values and output to stdout
call get_npar(npar)
allocate(dv(npar))
do i=1,npar
   call A3d(rad,theta,phi,i,dv(i))
   call get_par(i,par)
   print*,par,100.0*dv(i)
enddo

end program A3d_relative


