!-----------------------------------------------------------
module module_A3d
!-----------------------------------------------------------
  use def_gparam
  use earth_modele
  implicit none
  public :: init_A3d,A3d,A3dsingle,init_A3d_reg,A3d_full,get_npar &
       ,get_par,get_mean,get_rad_range
  private

  real(DP), dimension(:), allocatable :: mdl,kntrad,aknot,oknot &
       ,aknotreg,oknotreg,disc_rad,aknot2,oknot2
  real(DP), dimension(:,:,:), allocatable :: y2
  integer, parameter :: MAXPAR=2
  integer :: npar,ndisc,nknotreg,nvsplreg,nb_disc
  integer, dimension(MAXPAR) :: nknotA1,nknotA2
  integer, dimension(:), allocatable :: level,NBF,reg,indexreg &
       ,vindexreg,levelreg,disc_index,level2
  character, dimension(:), allocatable :: parblock
  character(len=52), dimension(:,:), allocatable:: name_rec_tab
  integer :: rang
  integer :: surface,NBPARAM
  logical :: vert_reg_exist=.false.,model1D_exist=.false.
  logical :: modele1D_read=.false.
  logical :: hknots2_exist=.false.,unconformal=.false.
  logical :: interp=.false.,TI_interp=.false.,first_mean=.true.
  character(len=20), parameter :: A3d_dat            ='A3d.dat'
  character(len=20), parameter :: hknots_dat         ='hknots.dat'
  character(len=20), parameter :: hknots2_dat        ='hknots2.dat'
  character(len=20), parameter :: vert_reg_dat       ='vert_reg.dat'
  character(len=20), parameter :: model1Dfile        ='model1D.dat'
  type(modele) :: modele1D

!adel is used by various subroutines
  integer, parameter :: nadel=7
  real(DP), dimension(nadel), parameter :: adel=(/63.4, 31.7, 15.8, 7.9, 3.96, 1.98, 0.99/)
  real(DP), dimension(nadel) :: spmean

!-----------------------------------------------------------
  contains
!-----------------------------------------------------------

!---------------------------------------------------
  subroutine init_A3d(rang_)
!now allows for one non-conformal grid
!---------------------------------------------------
    integer, intent(in) :: rang_

    integer :: unit1=301,unit2=302,dum1,dum2,i,j,k,n,mdim
    real :: dum3

    rang=rang_

    inquire(file=model1Dfile,exist=model1D_exist)
    if(model1D_exist) then
       call read_modele(model1Dfile,modele1D)
       modele1D_read=.true.
    endif
   
    open(unit1,file=hknots_dat,status='old')
    read(unit1,*) nknotA2(1)
    if(allocated(oknot).or.allocated(aknot).or.allocated(level)) then
       print*,'A3d already initiated'
       return
    endif
    allocate(oknot(nknotA2(1)),aknot(nknotA2(1)),level(nknotA2(1)))

    !read in knot locations
    j=1
    do i=1,nknotA2(1)
       read(unit1,*) oknot(i),aknot(i),level(i)
    enddo
    close(unit1)

    inquire(file=hknots2_dat,exist=hknots2_exist)
    if(hknots2_exist) then
       open(unit1,file=hknots2_dat,status='old')
       read(unit1,*) nknotA2(2)
       if(allocated(oknot2).or.allocated(aknot2).or.allocated(level2)) then
          print*,'A3d already initiated'
          return
       endif
       allocate(oknot2(nknotA2(2)),aknot2(nknotA2(2)),level2(nknotA2(2)))

       !read in knot locations
       j=1
       do i=1,nknotA2(2)
          read(unit1,*) oknot2(i),aknot2(i),level2(i)
       enddo
       close(unit1)
    else
       nknotA2(2)=nknotA2(1)
    endif

    !open model file and read in model
    open(unit2,file=A3d_dat,status='old')
    read(unit2,*) npar
    NBPARAM=npar
    if(npar>MAXPAR) &
         stop 'npar greater than MAXPAR'
    !for now npar should be 1
    !    if (npar /= 1) then
    !       if (rang==0) print*,'Assuming first par is iso Vs'
    !    endif
    allocate(parblock(npar))
    do i=1,npar
       read(unit2,*) dum2,nknotA1(i),parblock(i)
       if(i>1.and.nknotA1(i)/=nknotA1(1)) then
          stop 'Inconsistent A1 splines between parameters'
       endif
       if(dum2/=nknotA2(i)) &
            stop 'Inconsistent hknots.dat and A3d.dat'
       if(i>1.and.nknotA2(i)/=nknotA2(1)) then
          unconformal=.true.
          if(.not.hknots2_exist) &
               stop 'unconformal grid requires hknots2.dat'
       endif
    enddo
    read(unit2,*) ndisc 
    surface=0
    if (ndisc>0) then
       surface=ndisc
       if(unconformal) &
            print*,'discontinuities assumed same grid as first par'
    endif
    allocate(kntrad(nknotA1(1)))
    read(unit2,*) (kntrad(i),i=1,nknotA1(1))
    n=0
    mdim=0
    do i=1,npar
       mdim=mdim+nknotA1(i)*nknotA2(i)
    enddo
    mdim=mdim+ndisc*nknotA2(1)
    allocate(mdl(mdim))
    do i=1,npar
       do j=1,nknotA1(i)
          read(unit2,*) (mdl(k+n),k=1,nknotA2(i))
          n=n+nknotA2(i)
       enddo
    enddo
    do i=1,ndisc
      read(unit2,*) (mdl(k+n),k=1,nknotA2(1))
      n=n+nknotA2(1)
    enddo

    if(n/=mdim) stop 'init_A3d dimension error'
    close(unit2)
!---------------------------------------------------
  end subroutine init_A3d
!---------------------------------------------------
!---------------------------------------------------
  subroutine init_A3d_reg(rang_)
!no non-conformal grid
!---------------------------------------------------
    integer, intent(in) :: rang_

    integer :: unit1=301,unit2=302,unit3=303,dum1,dum2,i,j,k,n,mdim
    real :: dum3

    rang=rang_

    inquire(file=model1Dfile,exist=model1D_exist)
    if(model1D_exist) then
       call read_modele(model1Dfile,modele1D)
       modele1D_read=.true.
    endif
   
    open(unit1,file=hknots_dat,status='old')
    read(unit1,*) nknotA2(1)
    if(allocated(oknot).or.allocated(aknot).or.allocated(level)) then
       print*,'A3d already initiated'
       return
    endif
    allocate(oknot(nknotA2(1)),aknot(nknotA2(1)),level(nknotA2(1)))
    !additional regional indices
    if(allocated(reg).or.allocated(indexreg)) then
       print*,'A3d regional already initiated'
       return
    endif
    allocate(reg(nknotA2(1)),indexreg(nknotA2(1)))

    !read in knot locations
    j=1
    do i=1,nknotA2(1)
       read(unit1,*) oknot(i),aknot(i),level(i),reg(i)
       if (reg(i)==1) then
          indexreg(i)=j
          j=j+1
       else
          indexreg(i)=9999
       endif
    enddo
    nknotreg=j-1
    close(unit1)

    allocate(oknotreg(nknotreg),aknotreg(nknotreg),levelreg(nknotreg))
    do i=i,nknotA2(1)
       if(reg(i)==1) then
          oknotreg(indexreg(i))=oknot(i)
          aknotreg(indexreg(i))=aknot(i)
          levelreg(indexreg(i))=level(i)
       endif
    enddo

    !check if vert_reg_dat exists
    inquire(file=vert_reg_dat,exist=vert_reg_exist)
    if(vert_reg_exist) then
       open(unit3,file=vert_reg_dat,status='old')
       read(unit3,*) nknotA1(1)
       allocate(vindexreg(nknotA1(1)),kntrad(nknotA1(1)))
       nvsplreg=0
       do i=1,nknotA1(1)
          read(unit3,*) kntrad(i),dum1
          if (dum1==1) then
             nvsplreg=nvsplreg
             vindexreg(i)=nvsplreg
          else
             vindexreg(i)=9999
          endif
       enddo
       close(unit3)
    endif
       
    
    !open model file and read in model 
    !model is assumed to be global model
    open(unit2,file=A3d_dat,status='old')
    read(unit2,*) npar
    NBPARAM=npar

    if (npar /= 1) then
       if (rang==0) print*,'Assuming first par is iso Vs'
    endif
    allocate(parblock(npar))
    do i=1,npar
       read(unit2,*) dum1,dum2,parblock(i)
       if(i==1.and..not.vert_reg_exist) nknotA1(1)=dum2
       if(dum1/=nknotA2(1).or.dum2/=nknotA1(1)) then
          stop 'Inconsistent splines between parameters'
       endif
       if(i>1) then
          nknotA1(i)=nknotA1(1)
          nknotA2(i)=nknotA2(1)
       endif
    enddo
    read(unit2,*) ndisc 
    surface=0
    if (ndisc>0) surface=ndisc
    if(.not.vert_reg_exist) then
       allocate(kntrad(nknotA1(1)))
       read(unit2,*) (kntrad(i),i=1,nknotA1(1))
    else
       read(unit2,*) !should error check here
    endif
    n=0
    mdim=npar*nknotA1(1)*nknotA2(1)+ndisc*nknotA2(1)
    allocate(mdl(mdim))
    do i=1,npar
       do j=1,nknotA1(1)
          read(unit2,*) (mdl(k+n),k=1,nknotA2(1))
          n=n+nknotA2(1)
       enddo
    enddo
    do i=1,ndisc
      read(unit2,*) (mdl(k+n),k=1,nknotA2(1))
      n=n+nknotA2(1)
    enddo

    if(n/=mdim) stop 'init_A3d_reg dimension error'
    close(unit2)
!---------------------------------------------------
  end subroutine init_A3d_reg
!---------------------------------------------------
!---------------------------------------------------
  subroutine get_npar(npar_)
!---------------------------------------------------
    use def_gparam
    integer, intent(out) :: npar_

    npar_=npar
!---------------------------------------------------
  end subroutine get_npar
!---------------------------------------------------
!---------------------------------------------------
  subroutine get_par(n,par)
!---------------------------------------------------
    use def_gparam
    integer, intent(in) :: n
    character, intent(out) :: par

    if (n>npar) stop 'get_par: dimension error'
    par=parblock(n)
!---------------------------------------------------
  end subroutine get_par
!---------------------------------------------------

!---------------------------------------------------
  subroutine get_rad_range(rad1,rad2)
!---------------------------------------------------
!assumes only one A1 range
    real(DP), intent(out) :: rad1,rad2
    
    rad1=kntrad(1)
    rad2=kntrad(nknotA1(1))
!---------------------------------------------------
  end subroutine get_rad_range
!---------------------------------------------------


!---------------------------------------------------
  subroutine A3d(r,theta,phi,par,dv)
!---------------------------------------------------
!currently only returns one dv value 
    use def_gparam
    real(DP), intent(in ) :: r,theta,phi
    integer,  intent(in ) :: par
    real(DP), intent(out) :: dv

    integer :: jump,effnknot,i,j,k
    real(DP) :: lat,lon,del,dr,adel1
    real(DP), dimension(:), allocatable :: dh
    integer, dimension(:), allocatable :: kindex
    !real(DP), dimension(7) :: adel

    !adel=(/63.4, 31.7, 15.8, 7.9, 3.96, 1.98, 0.99/)
    if(unconformal.and.npar/=2) &
         stop 'A3d: unconformal grid only supported for npar 2'
    if(par<=npar) then
       if(r>kntrad(nknotA1(1)).or.r<kntrad(1)) then
          dv=0.0
       else
          jump=(par-1)*nknotA1(1)*nknotA2(1)
          allocate(dh(nknotA2(par)),kindex(nknotA2(par)))
          
          lat=90-rad2deg*theta
          lon=rad2deg*phi
          effnknot=0
          do i=1,nknotA2(par)
             if(unconformal.and.npar==2.and.par==2) then
                del=getdel(lat,lon,aknot2(i),oknot2(i))
                adel1=adel(level2(i))
             else
                del=getdel(lat,lon,aknot(i),oknot(i))
                adel1=adel(level(i))
             endif
             if(del<=adel1*2.0) then
                effnknot=effnknot+1
                kindex(effnknot)=i
                dh(effnknot)=spbsp(del,adel1)
             endif
          enddo
          
          dv=0.0
          do i=1,nknotA1(par)
             call fspl(i,nknotA1(par),kntrad,r,dr)
             do j=1,effnknot
                dv=dv+dr*dh(j)*mdl(jump+kindex(j)+nknotA2(par)*(i-1))
             enddo
          enddo
          deallocate(dh,kindex)
       endif
    else if (par<=npar+ndisc) then !discontinuties
       jump=npar*nknotA1(1)*nknotA2(1)
       if(unconformal) &
            jump=nknotA1(1)*nknotA2(1)+nknotA1(2)+nknotA2(2)
       jump=jump+(par-npar-1)*nknotA2(1)
       allocate(dh(nknotA2(1)),kindex(nknotA2(1)))
          
       lat=90-rad2deg*theta
       lon=rad2deg*phi
       effnknot=0
       do i=1,nknotA2(1)
          del=getdel(lat,lon,aknot(i),oknot(i))
          if(del<=adel(level(i))*2.0) then
             effnknot=effnknot+1
             kindex(effnknot)=i
             dh(effnknot)=spbsp(del,adel(level(i)))
          endif
       enddo
          
       dv=0.0
       do j=1,effnknot
          dv=dv+dh(j)*mdl(jump+kindex(j))
       enddo
       deallocate(dh,kindex)
    else
       dv=0.0;
    endif
       
!---------------------------------------------------
  end subroutine A3d
!---------------------------------------------------

!---------------------------------------------------
  subroutine A3d_full(r,theta,phi,ifanis,vs,vp,rho,Qs,xi,fi,eta)
!---------------------------------------------------
!returns isotropic vs,vp, and rho assuming scaling
!dlnVs/dlnVp=2 dlnVs/dlnrho=3
    use def_gparam
    real(DP), intent(in ) :: r,theta,phi
    integer,  intent(in ) :: ifanis
    real(DP), intent(out) :: vs,vp,rho,Qs
    real(DP), optional, intent(out) :: xi,fi,eta

    integer :: jump,effnknot,i,j,k
    real(DP) :: lat,lon,del,dr,dv,AA,CC,FF,LL,NN,eta1,adel1
    real(DP), dimension(:), allocatable :: dh
    integer, dimension(:), allocatable :: kindex
    !real(DP), dimension(5) :: adel

    !adel=(/63.4, 31.7, 15.8, 7.9, 3.95/)

    if(.not.model1D_exist) &
         stop 'no 1D model file'

    if(ifanis==1) then
       if(.not.(present(xi).and.present(fi).and.present(eta))) &
            stop 'A3d_full: ifanis inconsistent'
    endif

    call get_1Dmodel_TI(AA,CC,FF,LL,NN,rho,Qs,r*1000.0_DP)

    eta1=FF/(AA-2.*LL)
    !Voigt average
    vp=sqrt((3.*CC+(8.+4.*eta1)*AA+8.*(1.-eta1)*LL)/(15.*rho))
    vs=sqrt((CC+(1.-2.*eta1)*AA+(6.+4.*eta1)*LL+5.*NN)/(15.*rho))

    if(ifanis==1) then
       eta=eta1
       xi=NN/LL
       fi=CC/AA
    endif


    if(r>kntrad(nknotA1(1)).or.r<kntrad(1)) then
       dv=0.0
    else
       jump=0
       allocate(dh(nknotA2(1)),kindex(nknotA2(1)))
       
       lat=90-rad2deg*theta
       lon=rad2deg*phi
       effnknot=0
       do i=1,nknotA2(1)
          del=getdel(lat,lon,aknot(i),oknot(i))
          if(del<=adel(level(i))*2.0) then
             effnknot=effnknot+1
             kindex(effnknot)=i
             dh(effnknot)=spbsp(del,adel(level(i)))
          endif
       enddo
          
       dv=0.0
       do i=1,nknotA1(1)
          call fspl(i,nknotA1(1),kntrad,r,dr)
          do j=1,effnknot
             dv=dv+dr*dh(j)*mdl(jump+kindex(j)+nknotA2(1)*(i-1))
          enddo
       enddo
       deallocate(dh,kindex)
    endif
 
    vs=vs+dv*vs
    vp=vp+0.5_DP*dv*vp
    rho=rho+0.33333_DP*dv*rho
 
    if(ifanis==1) then
       if(npar<2) then !no xi
          dv=0.
       else if (npar==2) then
          jump=nknotA2(1)*nknotA1(1)
          allocate(dh(nknotA2(2)),kindex(nknotA2(2)))
       
          lat=90-rad2deg*theta
          lon=rad2deg*phi
          effnknot=0
          do i=1,nknotA2(2)
             if(unconformal) then
                del=getdel(lat,lon,aknot2(i),oknot2(i))
                adel1=adel(level2(i))
             else
                del=getdel(lat,lon,aknot(i),oknot(i))
                adel1=adel(level(i))
             endif
             if(del<=adel1*2.0) then
                effnknot=effnknot+1
                kindex(effnknot)=i
                dh(effnknot)=spbsp(del,adel1)
             endif
          enddo
          
          dv=0.0
          do i=1,nknotA1(2)
             call fspl(i,nknotA1(2),kntrad,r,dr)
             do j=1,effnknot
                dv=dv+dr*dh(j)*mdl(jump+kindex(j)+nknotA2(2)*(i-1))
             enddo
          enddo
          deallocate(dh,kindex)
       else
          stop 'npar > 2 not supported yet'
       endif

       xi=xi+dv*xi
       fi=fi-1.5*dv*fi
       eta=eta-2.5*dv*eta
    endif

!---------------------------------------------------
  end subroutine A3d_full
!---------------------------------------------------

!---------------------------------------------------
  subroutine A3dsingle(r,theta,phi,par,if,dv)
!get amplitude of 1 model coefficient at a point
!---------------------------------------------------
    use def_gparam
    real(DP), intent(in ) :: r,theta,phi
    integer,  intent(in ) :: par,if
    complex, dimension(:), intent(out) :: dv

    integer :: num,A1index,A2index,i,j
    real(DP) :: lat,lon,del,dr,dh
    !real(DP), dimension(5) :: adel

!temporary fix until A3d is ready to deal with anis
!    if (NBPARAM/=1) then
!       stop 'A3d format only handles isotropic S to date'
!    endif
    !adel=(/63.4, 31.7, 15.8, 7.9, 3.9/)

    if (nknotreg==-1) then
       if (par==1) then
          do num=1,NBPARAM
             if(r>kntrad(nknotA1(1)).or.r<kntrad(1)) then
                dv(num)=cmplx(0.0)
             else
                A1index=ceiling(real(if)/real(nknotA2(1)))
                A2index=if-(A1index-1)*nknotA2(1)
                lat=90-rad2deg*theta
                lon=rad2deg*phi
                del=getdel(lat,lon,aknot(A2index),oknot(A2index))
                if (del<=adel(level(A2index))*2.0) then
                   dh=spbsp(del,adel(level(A2index)))
                   call fspl(A1index,nknotA1(1),kntrad,r,dr)
                   dv(num)=cmplx(dr*dh)
                else
                   dv(num)=cmplx(0.0)
                endif
             endif
          enddo
       else if (par < ndisc+2) then !surface terms
          if (surface==0) stop 'module_A3d: no discontinuities in model'
          A1index=ceiling(real(if)/real(nknotA2(1))) !all disc param. equiv.
          if (A1index.ne.(par-2)) then
             dv(1)=cmplx(0.0)
          else  
             A2index=if-(A1index-1)*nknotA2(1)
             lat=90-rad2deg*theta
             lon=rad2deg*phi
             del=getdel(lat,lon,aknot(A2index),oknot(A2index))
             if (del<=adel(level(A2index))*2.0) then
                dh=spbsp(del,adel(level(A2index)))
                dv(1)=cmplx(dh)
             else
                dv(1)=cmplx(0.0)
             endif
          endif
       else
          stop 'A3dsingle: unknown par value'   
       endif
    else !regional case
       if (par==1) then
          do num=1,NBPARAM
             if(r>kntrad(nknotA1(1)).or.r<kntrad(1)) then
                dv(num)=cmplx(0.0)
             else
                A1index=ceiling(real(if)/real(nknotreg))
                i=1
                do while (i<=nknotA2(1))
                   if (reg(i)==1) then
                      if (indexreg(i)==(if-(A1index-1)*nknotreg)) then
                         A2index=i
                         i=nknotA2(1)
                      endif
                   endif
                   i=i+1
                enddo
                lat=90-rad2deg*theta
                lon=rad2deg*phi
                del=getdel(lat,lon,aknot(A2index),oknot(A2index))
                if (del<=adel(level(A2index))*2.0) then
                   dh=spbsp(del,adel(level(A2index)))
                   if(vert_reg_exist) then
                      do i=1,nknotA1(1)
                         if(vindexreg(i)==A1index) j=i
                      enddo
                      call fspl(j,nknotA1(1),kntrad,r,dr)
                   else
                      call fspl(A1index,nknotA1(1),kntrad,r,dr)
                   endif
                   dv(num)=cmplx(dr*dh)
                else
                   dv(num)=cmplx(0.0)
                endif
             endif
          enddo
       else if (par < ndisc+2) then !surface terms
          if (surface==0) stop 'module_A3d: no discontinuities in model'
          A1index=ceiling(real(if)/real(nknotreg)) 
          if (A1index.ne.(par-2)) then
             dv(1)=cmplx(0.0)
          else  
             i=1
             do while (i<=nknotA2(1))
                if (reg(i)==1) then
                   if (indexreg(i)==(if-(A1index-1)*nknotreg)) then
                      A2index=i
                      i=nknotA2(1)
                   endif
                endif
                i=i+1
             enddo
             lat=90-rad2deg*theta
             lon=rad2deg*phi
             del=getdel(lat,lon,aknot(A2index),oknot(A2index))
             if (del<=adel(level(A2index))*2.0) then
                dh=spbsp(del,adel(level(A2index)))
                dv(1)=cmplx(dh)
             else
                dv(1)=cmplx(0.0)
             endif
          endif
       else
          stop 'A3dsingle: unknown par value'   
       endif
    endif
     
!---------------------------------------------------
  end subroutine A3dsingle
!---------------------------------------------------

!---------------------------------------------------
  real(DP) function getdel(a0,o0,a,o)
!---------------------------------------------------
    use def_gparam
    real(DP) :: a0,o0,a,o

    real(DP) :: q0,sq0,cq0,q,sq,cq,ff,sff,cff,arg

    q0=(90.0-a0)*deg2rad;
    sq0=sin(q0);
    cq0=cos(q0);

    q=(90.0-a)*deg2rad;
    sq=sin(q);
    cq=cos(q);
    
    ff=(o-o0)*deg2rad;
    sff=sin(ff);
    cff=cos(ff);

    arg=cq*cq0+sq*sq0*cff;
    if(arg > 1.) arg= 1.;
    if(arg < -1.) arg=-1.;
    getdel=rad2deg*acos(arg);
!---------------------------------------------------
  end function getdel
!---------------------------------------------------

!---------------------------------------------------
  real(DP) function spbsp(hdel,ahdel)
!---------------------------------------------------
    real(DP) :: hdel,ahdel

    if(hdel < ahdel) then
       spbsp=0.75*(hdel/ahdel)*(hdel/ahdel)*(hdel/ahdel)-1.5*(hdel/ahdel)*(hdel/ahdel)+1.
    else if(hdel <= ahdel*2.) then
       spbsp=0.25*(2.-hdel/ahdel)*(2.-hdel/ahdel)*(2.-hdel/ahdel)
       if(spbsp < 0) spbsp=0.0 
    else
       spbsp=0.0
    endif
!---------------------------------------------------
  end function spbsp
!---------------------------------------------------

!---------------------------------------------------
  real(DP) function mean_spbsp(ahdel)
!---------------------------------------------------
! computes the mean value of a spbsp spline on the unit sphere
! determined by centering a spline on the north pole and then 
! computing integral average value:
! 1/(4*pi) int_Omega f(theta) sin(theta) dOmega
! where Omega is the solid abgle for the unit sphere, and f is 
! the funtion defined in spbsp
    real(DP), intent(in) :: ahdel

    real(DP) :: ahdelrad,ahdelrad2,ahdelrad3

    ahdelrad=deg2rad*ahdel
    ahdelrad2=ahdelrad*ahdelrad
    ahdelrad3=ahdelrad2*ahdelrad

    mean_spbsp=0.5d0*ahdelrad3+1.5d0*ahdelrad-3.d0*sin(ahdelrad) &
         +0.75d0*sin(2.d0*ahdelrad)
    mean_spbsp=mean_spbsp/ahdelrad3

!---------------------------------------------------
  end function mean_spbsp
!---------------------------------------------------

!---------------------------------------------------
  subroutine get_mean(r,par,mean)
!---------------------------------------------------
    real(DP), intent(in ) :: r
    integer,  intent(in ) :: par
    real(DP), intent(out) :: mean
    
    integer  :: i,j,jump
    real(DP) :: dr

    if(first_mean) then
       do i=1,nadel
          spmean(i)=mean_spbsp(adel(i))
       enddo
       first_mean=.false.
    endif

    if(par<=npar) then
       if(r>kntrad(nknotA1(1)).or.r<kntrad(1)) then
          mean=0.0
       else
          if(unconformal.and.npar/=2) &
               stop 'A3d: unconformal grid only supported for npar 2'
          jump=(par-1)*nknotA1(1)*nknotA2(1)
          mean=0.0
          do i=1,nknotA1(par)
             call fspl(i,nknotA1(par),kntrad,r,dr)
             if(dr/=0.0) then
                if(unconformal.and.par==2) then
                   do j=1,nknotA2(par)
                      mean=mean+dr*spmean(level2(j)) &
                           *mdl(jump+(i-1)*nknotA2(par)+j)
                   enddo
                else
                   do j=1,nknotA2(par)
                      mean=mean+dr*spmean(level(j)) &
                           *mdl(jump+(i-1)*nknotA2(par)+j)
                   enddo
                endif
             endif
          enddo
       endif
    else if(par<=npar+ndisc) then !discontinuity
       jump=0
       do i=1,npar
          jump=jump+nknotA1(i)*nknotA2(i)
       enddo
       jump=jump+(par-npar-1)*nknotA2(1) !all disc assumed same as first par

       mean=0.0
       do i=1,nknotA2(1)
          mean=mean+spmean(level(i))*mdl(jump+i)
       enddo
    else
       print*,'Invalid par number:',par
       mean=0.0
    endif

!-----------------------------------------------------------------
  end subroutine get_mean
!-----------------------------------------------------------------

!-----------------------------------------------------------------
      subroutine get_1Dmodel(rho,Vs,Vp,Qs,rad)
!-----------------------------------------------------------------
        use def_gparam
        use module_spline
        implicit none
        real(DP), intent(in ) :: rad
        real(DP), intent(out) :: rho,Vs,Vp,Qs
!
        integer :: i,k,ka,kb,nd,si,kas,kbs
!        real(DP), dimension(nkT,nb_discT+1,4)   :: y2
        real(DP)        :: yp1,ypn
!d'abord on cherche l'index de de depart: (si=start_index)
!        do i=1,modele1D%nbcou
!           if (abs(modele1D%r(i)-rayT(1))/rayT(1)< 1.E-7_DP) si=i
!        enddo
!        si=si-1
!        print*,'start_index=',si
!
        if (modele1D_read.and..not.interp) then
           interp=.true.
           nb_disc=0
           do i=1,modele1D%nbcou
              if (abs(modele1D%r(i)-modele1D%r(i+1))<1e-8) &
                   nb_disc=nb_disc+1
           enddo
           allocate(disc_index(0:nb_disc+1),disc_rad(0:nb_disc+1))
           disc_index(0)=0
           disc_index(nb_disc+1)=modele1D%nbcou
           disc_rad(0)  =modele1D%r(1)
           disc_rad(nb_disc+1) =modele1D%r(modele1D%nbcou)
           k=0
           do i=1,modele1D%nbcou-1
              if (abs(modele1D%r(i)-modele1D%r(i+1))<1e-8) then
                 k=k+1
                 disc_index(k)=i
                 disc_rad  (k)=modele1D%r(i)
              endif
           enddo
           
           if(.not.allocated(y2)) allocate(y2(modele1D%nbcou,nb_disc+1,4))
           
           do nd=1,nb_disc+1
              ka=disc_index(nd-1)+1 
              kb=disc_index(nd)     
!
              yp1=(modele1D%rho(ka+1)-modele1D%rho(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%rho(kb  )-modele1D%rho(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%rho(ka:kb),yp1,ypn,y2(ka:kb,nd,1))                 
!
              yp1=(modele1D%vpv(ka+1)-modele1D%vpv(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%vpv(kb  )-modele1D%vpv(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%vpv(ka:kb),yp1,ypn,y2(ka:kb,nd,2))
!                 
              yp1=(modele1D%vsv(ka+1)-modele1D%vsv(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%vsv(kb  )-modele1D%vsv(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%vsv(ka:kb),yp1,ypn,y2(ka:kb,nd,3))                 
!
              yp1=(modele1D%qshear(ka+1)-modele1D%qshear(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%qshear(kb  )-modele1D%qshear(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%qshear(ka:kb),yp1,ypn,y2(ka:kb,nd,4))                 
!
           enddo
        endif
        nd=locate(disc_rad,nb_disc,rad)
        ka=disc_index(nd-1)+1 
        kb=disc_index(nd)     
        rho=splint(modele1D%r(ka:kb),modele1D%rho(ka:kb)   ,y2(ka:kb,nd,1),rad)
        vp =splint(modele1D%r(ka:kb),modele1D%vpv(ka:kb)   ,y2(ka:kb,nd,2),rad)
        vs =splint(modele1D%r(ka:kb),modele1D%vsv(ka:kb)   ,y2(ka:kb,nd,3),rad)
        qs =splint(modele1D%r(ka:kb),modele1D%qshear(ka:kb),y2(ka:kb,nd,4),rad)

!-----------------------------------------------------------------
      end subroutine get_1Dmodel
!-----------------------------------------------------------------
!-----------------------------------------------------------------
      subroutine get_1Dmodel_TI(AA,CC,FF,LL,NN,rho,Qs,rad)
!-----------------------------------------------------------------
        use def_gparam
        use module_spline
        implicit none
        real(DP), intent(in ) :: rad
        real(DP), intent(out) :: AA,CC,FF,LL,NN,rho,Qs
!
        integer :: i,k,ka,kb,nd,si,kas,kbs
!        real(DP), dimension(nkT,nb_discT+1,7)   :: y2
        real(DP)        :: yp1,ypn
!d'abord on cherche l'index de de depart: (si=start_index)
!        do i=1,modele1D%nbcou
!           if (abs(modele1D%r(i)-rayT(1))/rayT(1)< 1.E-7_DP) si=i
!        enddo
!        si=si-1
!        print*,'start_index=',si
!
        if (modele1D_read.and..not.interp) then
           interp=.true.
           nb_disc=0
           do i=1,modele1D%nbcou
              if (abs(modele1D%r(i)-modele1D%r(i+1))<1e-8) &
                   nb_disc=nb_disc+1
           enddo
           allocate(disc_index(0:nb_disc+1),disc_rad(0:nb_disc+1))
           disc_index(0)=0
           disc_index(nb_disc+1)=modele1D%nbcou
           disc_rad(0)  =modele1D%r(1)
           disc_rad(nb_disc+1) =modele1D%r(modele1D%nbcou)
           k=0
           do i=1,modele1D%nbcou-1
              if (abs(modele1D%r(i)-modele1D%r(i+1))<1e-8) then
                 k=k+1
                 disc_index(k)=i
                 disc_rad  (k)=modele1D%r(i)
              endif
           enddo
           
           if(.not.allocated(y2)) allocate(y2(modele1D%nbcou,nb_disc+1,7))
           
           do nd=1,nb_disc+1
              ka=disc_index(nd-1)+1 
              kb=disc_index(nd)     
!
              yp1=(modele1D%rho(ka+1)-modele1D%rho(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%rho(kb  )-modele1D%rho(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%rho(ka:kb),yp1,ypn,y2(ka:kb,nd,1))                 
!
              yp1=(modele1D%vpv(ka+1)-modele1D%vpv(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%vpv(kb  )-modele1D%vpv(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%vpv(ka:kb),yp1,ypn,y2(ka:kb,nd,2))
!                 
              yp1=(modele1D%vsv(ka+1)-modele1D%vsv(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%vsv(kb  )-modele1D%vsv(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%vsv(ka:kb),yp1,ypn,y2(ka:kb,nd,3))                 
!
              yp1=(modele1D%qshear(ka+1)-modele1D%qshear(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%qshear(kb  )-modele1D%qshear(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%qshear(ka:kb),yp1,ypn,y2(ka:kb,nd,4))                 
!
              yp1=(modele1D%vph(ka+1)-modele1D%vph(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%vph(kb  )-modele1D%vph(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%vph(ka:kb),yp1,ypn,y2(ka:kb,nd,5))
!                 
              yp1=(modele1D%vsh(ka+1)-modele1D%vsh(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%vsh(kb  )-modele1D%vsh(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%vsv(ka:kb),yp1,ypn,y2(ka:kb,nd,6))
!                 
              yp1=(modele1D%eta(ka+1)-modele1D%eta(ka  ))/(modele1D%r(ka+1)-modele1D%r(ka  ))
              ypn=(modele1D%eta(kb  )-modele1D%eta(kb-1))/(modele1D%r(kb  )-modele1D%r(kb-1))
              call spline(modele1D%r(ka:kb),modele1D%eta(ka:kb),yp1,ypn,y2(ka:kb,nd,7))
!                 
           enddo
        endif
        nd=locate(disc_rad,nb_disc,rad)
        ka=disc_index(nd-1)+1 
        kb=disc_index(nd)     
        rho=splint(modele1D%r(ka:kb),modele1D%rho(ka:kb),y2(ka:kb,nd,1) &
             ,rad)
        AA=splint(modele1D%r(ka:kb),modele1D%vph(ka:kb),y2(ka:kb,nd,5) &
             ,rad)**2*rho
        CC=splint(modele1D%r(ka:kb),modele1D%vpv(ka:kb),y2(ka:kb,nd,2) &
             ,rad)**2*rho
        LL=splint(modele1D%r(ka:kb),modele1D%vsv(ka:kb),y2(ka:kb,nd,3) &
             ,rad)**2*rho
        NN=splint(modele1D%r(ka:kb),modele1D%vsh(ka:kb),y2(ka:kb,nd,6) &
             ,rad)**2*rho
        FF=splint(modele1D%r(ka:kb),modele1D%eta(ka:kb),y2(ka:kb,nd,7) &
             ,rad)*(AA-2._DP*LL)
        qs=splint(modele1D%r(ka:kb),modele1D%qshear(ka:kb) &
                   ,y2(ka:kb,nd,4),rad)
          
!-----------------------------------------------------------------
      end subroutine get_1Dmodel_TI
!-----------------------------------------------------------------

!-----------------------------------------------------------------
      integer function locate(inter,n,x)
!-----------------------------------------------------------------
         use def_gparam
         implicit none      
         integer, intent(in) :: n
         real(DP), dimension(0:n+1), intent(in) :: inter
         real(DP), intent(in) :: x         
!
         integer :: i
         i=0
         if (abs (x-inter(0))/max(x,inter(0)) <1.E-10_DP) then
            i=0
         else if ( abs (x-inter(n+1))/max(x,inter(n+1))  <1.E-10_DP) then
            i=n
         else
            do while (  x < inter(i) .or. x > inter(i+1) )
               i=i+1
               if (i > n) then
                  print*,'i=',i,'n=',n
                  print*,'x=',x
                  print*,'inter=',inter
                  stop 'locate: failed to locate x in inter!'
               endif
            enddo
         endif
         locate=i+1
!-----------------------------------------------------------------
      end function locate
!-----------------------------------------------------------------


!-----------------------------------------------------------
end module module_A3d
!-----------------------------------------------------------
